@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.009';
+our $VERSION = '0.010';
 
 use EV;
 use Time::HiRes;
@@ -92,6 +92,7 @@ sub new {
         debug               => delete $args->{debug} || 0,
         keep_timeline       => delete $args->{keep_timeline} || 0,
         pool_args           => $args,
+        is_running          => 0,
     }, $class;
 
     weaken($self->{storage});
@@ -149,6 +150,13 @@ sub request {
     $self->{connections}{$conn_id} = $conn;
 
     $self->_set_init_state($conn_id);
+
+    # if user fire new request in a callback we need to update stop_condition
+    my $stop_condition = $self->{stop_condition};
+    if ($stop_condition && $stop_condition->{all}) {
+        $stop_condition->{connections}{$conn_id} = 1;
+    }
+
     return $conn;
 }
 
@@ -165,9 +173,11 @@ sub drop {
     return delete $self->{connections}{$conn_id};
 }
 
-sub run      { shift->_run(0, @_)          }
-sub run_once { shift->_run(EV::RUN_ONCE)   }
-sub run_tick { shift->_run(EV::RUN_NOWAIT) }
+sub run         { shift->_run(0, @_)            }
+sub run_once    { shift->_run(EV::RUN_ONCE)     }
+sub run_tick    { shift->_run(EV::RUN_NOWAIT)   }
+sub is_running  { shift->{is_running}           }
+sub loop        { shift->{loop}                 }
 
 ################################################################################
 # Routines to manipulate connections (also user facing)
@@ -228,6 +238,8 @@ sub yahc_conn_url {
 sub _run {
     my ($self, $how, $until_state, @cs) = @_;
     die "YAHC: storage object is destroyed\n" unless $self->{storage};
+    die "YAHC: reentering run\n" if $self->{is_running};
+    $self->{is_running} = 1;
 
     if ($self->{pid} != $$) {
         _log_message('Reinitializing event loop after forking') if $self->{debug};
@@ -239,14 +251,15 @@ sub _run {
         my $until_state_str = _strstate($until_state);
         die "YAHC: unknown until_state $until_state\n" if $until_state_str =~ m/unknown/;
 
-        my @connections = (@cs == 0)
-                        ? values %{ $self->{connections} }
-                        : map { $self->{connections}{$_} || () }
-                          map { ref($_) eq 'HASH' ? $_->{id} : $_ } @cs;
+        my $is_all = (@cs == 0);
+        my @connections = $is_all ? values %{ $self->{connections} }
+                                  : map { $self->{connections}{$_} || () }
+                                    map { ref($_) eq 'HASH' ? $_->{id} : $_ } @cs;
 
         $self->{stop_condition} = {
-            expected_state => $until_state,
-            connections => { map { $_->{id} => 1 } grep { $_->{state} < $until_state } @connections },
+            all             => $is_all,
+            expected_state  => $until_state,
+            connections     => { map { $_->{id} => 1 } grep { $_->{state} < $until_state } @connections },
         };
     } else {
         delete $self->{stop_condition};
@@ -254,12 +267,17 @@ sub _run {
 
     my $loop = $self->{loop};
     $loop->now_update;
-    return $loop->run($how || 0) unless $self->{debug}; # shortcut
 
-    my $iterations = $loop->iteration;
-    _log_message('pid %d entering event loop%s', $$, ($until_state ? " with until state " . _strstate($until_state) : ''));
-    $loop->run($how || 0);
-    _log_message('pid %d exited from event loop after %d iterations', $$, $loop->iteration - $iterations);
+    if ($self->{debug}) {
+        my $iterations = $loop->iteration;
+        _log_message('pid %d entering event loop%s', $$, ($until_state ? " with until state " . _strstate($until_state) : ''));
+        $loop->run($how || 0);
+        _log_message('pid %d exited from event loop after %d iterations', $$, $loop->iteration - $iterations);
+    } else {
+        $loop->run($how || 0);
+    }
+
+    $self->{is_running} = 0;
 }
 
 sub _break {
@@ -292,8 +310,9 @@ sub _check_stop_condition {
 sub _set_init_state {
     my ($self, $conn_id) = @_;
 
-    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
-    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
+    my $conn = $self->{connections}{$conn_id};
+    my $watchers = $self->{watchers}{$conn_id};
+    return $self->_set_completed_state($conn_id) unless $conn && $watchers;
 
     $conn->{response} = { status_code => 0 };
     $conn->{state} = YAHC::State::INITIALIZED();
@@ -302,10 +321,9 @@ sub _set_init_state {
 
     my $continue = 1;
     while ($continue) {
-        if (my $w = delete $watchers->{io}) {
-            $w->stop;
-            shutdown($w->fh, 2);
-        }
+        delete $watchers->{io}; # implicit stop
+        my $fh = delete $watchers->{_fh};
+        $fh && shutdown($fh, 2), undef $fh;
 
         my $attempt = $conn->{attempt}++;
         if ($attempt > $conn->{retries}) {
@@ -342,15 +360,16 @@ sub _set_init_state {
 sub _set_wait_synack_state {
     my ($self, $conn_id, $sock, $ip, $host, $port) = @_;
 
-    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
-    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
+    my $conn = $self->{connections}{$conn_id};
+    my $watchers = $self->{watchers}{$conn_id};
+    return $self->_set_completed_state($conn_id) unless $conn && $watchers;
     _assert_state($conn, YAHC::State::INITIALIZED()) if $conn->{debug};
 
     $conn->{state} = YAHC::State::WAIT_SYNACK();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
     $self->_call_state_callback($conn, 'wait_synack_callback') if $conn->{has_wait_synack_callback};
 
-    my $wait_synack_cb = $self->_get_safe_wrapper($conn, sub {
+    my $wait_synack_cb = sub {
         my $sockopt = getsockopt($sock, SOL_SOCKET, SO_ERROR);
         if (!$sockopt) {
             _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to do getsockopt(): $!");
@@ -370,8 +389,9 @@ sub _set_wait_synack_state {
         $self->_call_state_callback($conn, 'connected_callback') if $conn->{has_connected_callback};
 
         $self->_set_write_state($conn_id);
-    });
+    };
 
+    $watchers->{_fh} = $sock;
     $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, $wait_synack_cb);
     $self->_check_stop_condition($conn) if $self->{stop_condition};
 }
@@ -379,38 +399,35 @@ sub _set_wait_synack_state {
 sub _set_write_state {
     my ($self, $conn_id) = @_;
 
-    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
-    my $watcher = $self->{watchers}{$conn_id}{io} or die "YAHC: no watchers for connection id $conn_id\n";
+    my $conn = $self->{connections}{$conn_id};
+    my $watchers = $self->{watchers}{$conn_id};
+    my $watcher = $watchers->{io};
+    return $self->_set_completed_state($conn_id) unless $conn && $watchers && $watcher;
     _assert_connected($conn) if $conn->{debug};
 
     $conn->{state} = YAHC::State::WRITING();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
     $self->_call_state_callback($conn, 'writing_callback') if $conn->{has_writing_callback};
 
-    my $fd = fileno($watcher->fh);
+    my $fd = fileno($watchers->{_fh});
     my $buf = _build_http_message($conn);
     my $length = length($buf);
 
     _register_in_timeline($conn, "sending request of $length bytes") if $conn->{keep_timeline};
 
-    my $write_cb = $self->_get_safe_wrapper($conn, sub {
+    my $write_cb = sub {
         my $wlen = POSIX::write($fd, $buf, $length);
 
         if (!defined $wlen || $wlen == 0) {
             return if $! == EWOULDBLOCK || $! == EAGAIN || $! == EINTR;
-
-            my $error = !defined $wlen
-                      ? "Failed to send TCP data: $!"
-                      : "Bizzare!!! syswrite returned 0"; # wlen == 0
-
-            _register_error($conn, YAHC::Error::WRITE_ERROR(), $error);
+            _register_error($conn, YAHC::Error::WRITE_ERROR(), "Failed to send TCP data: $!");
             $self->_set_init_state($conn_id);
         } else {
             substr($buf, 0, $wlen, '');
             $length -= $wlen;
             $self->_set_read_state($conn_id) if $length == 0;
         }
-    });
+    };
 
     $watcher->cb($write_cb);
     $watcher->events(EV::WRITE);
@@ -420,8 +437,10 @@ sub _set_write_state {
 sub _set_read_state {
     my ($self, $conn_id) = @_;
 
-    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
-    my $watcher = $self->{watchers}{$conn_id}{io} or die "YAHC: no watchers for connection id $conn_id\n";
+    my $conn = $self->{connections}{$conn_id};
+    my $watchers = $self->{watchers}{$conn_id};
+    my $watcher = $watchers->{io};
+    return $self->_set_completed_state($conn_id) unless $conn && $watchers && $watcher;
     _assert_connected($conn) if $conn->{debug};
 
     $conn->{state} = YAHC::State::READING();
@@ -432,9 +451,9 @@ sub _set_read_state {
     my $neck_pos = 0;
     my $decapitated = 0;
     my $content_length = 0;
-    my $fd = fileno($watcher->fh);
+    my $fd = fileno($watchers->{_fh});
 
-    my $read_cb = $self->_get_safe_wrapper($conn, sub {
+    my $read_cb = sub {
         my $rlen = POSIX::read($fd, my $b = '', TCP_READ_CHUNK);
 
         if (!defined $rlen || $rlen == 0) {
@@ -452,7 +471,7 @@ sub _set_read_state {
             $buf .= $b;
             if (!$decapitated && ($neck_pos = index($buf, "${CRLF}${CRLF}")) > 0) {
                 my $headers = _parse_http_headers($conn, substr($buf, 0, $neck_pos, ''));
-                if (not defined $headers) {
+                if (!defined $headers || !exists $headers->{'Content-Length'}) {
                     $self->_set_user_action_state($conn_id, YAHC::Error::RESPONSE_ERROR(), "unsupported HTTP reponse");
                     return;
                 }
@@ -470,7 +489,7 @@ sub _set_read_state {
                 $self->_set_user_action_state($conn_id);
             }
         }
-    });
+    };
 
     $watcher->cb($read_cb);
     $watcher->events(EV::READ);
@@ -512,7 +531,7 @@ sub _set_user_action_state {
     _register_in_timeline($conn, "after invoking callback state is %s", _strstate($state)) if $conn->{keep_timeline};
 
     if ($state == YAHC::State::INITIALIZED()) {
-        $self->_set_init_state($conn_id); # TODO eval
+        $self->_set_init_state($conn_id);
     } elsif ($state == YAHC::State::USER_ACTION() || $state == YAHC::State::COMPLETED()) {
         $self->_set_completed_state($conn_id);
     } else {
@@ -535,14 +554,11 @@ sub _set_completed_state {
     $conn->{state} = YAHC::State::COMPLETED();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{keep_timeline};
 
-    if (my $w = delete $watchers->{io}) {
-        $w->stop;
-        shutdown($w->fh, 2)
-    }
+    my $fh = $watchers->{_fh};
+    $fh && shutdown($fh, 2);
+    undef $watchers; # implicit stop
 
-    $_->stop foreach (values %{ $watchers || {} });
-
-    eval { $self->_check_stop_condition($conn) } if $self->{stop_condition};
+    $self->_check_stop_condition($conn) if $self->{stop_condition};
 }
 
 sub _build_socket_and_connect {
@@ -588,28 +604,25 @@ sub _set_until_state_timer {
     my ($self, $conn_id, $timeout_name, $state, $error_to_report) = @_;
 
     my $timer_name = $timeout_name . '_timer';
-    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
-    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
-
-    if (my $timer = delete $watchers->{$timer_name}) {
-        $timer->stop;
-    }
+    my $conn = $self->{connections}{$conn_id};
+    my $watchers = $self->{watchers}{$conn_id};
+    return $self->_set_completed_state($conn_id) unless $conn && $watchers;
 
+    delete $watchers->{$timer_name}; # implicit stop
     my $timeout = $conn->{request}{$timeout_name};
     return unless $timeout;
 
-    my $timer_cb = $self->_get_safe_wrapper($conn, sub {
+    my $timer_cb = sub {
         if ($conn->{state} < $state) {
             _register_error($conn, $error_to_report, "$timeout_name of %.3fs expired", $timeout);
             $self->_set_init_state($conn_id);
         } else {
             _register_in_timeline($conn, "delete $timer_name") if $conn->{keep_timeline};
         }
-    });
+    };
 
     _register_in_timeline($conn, "setting $timeout_name to %.3fs", $timeout) if $conn->{keep_timeline};
     $watchers->{$timer_name} = $self->{loop}->timer($timeout, 0, $timer_cb);
-    $watchers->{$timer_name}->priority(-2); # set lowest priority
 }
 
 ################################################################################
@@ -691,19 +704,6 @@ sub _call_state_callback {
     # $self->{loop}->now_update; # XXX expect state callbacks to be small
 }
 
-sub _get_safe_wrapper {
-    my ($self, $conn, $sub) = @_;
-    return sub { eval {
-        $sub->(@_);
-        1;
-    } or do {
-        my $error = $@ || 'zombie error';
-        _register_error($conn, YAHC::Error::INTERNAL_ERROR(), "Exception callback: $error");
-        warn "YAHC: exception in callback: $error";
-        $self->_set_completed_state($conn->{id});
-    }};
-}
-
 sub _register_in_timeline {
     my ($conn, $format, @arguments) = @_;
     my $event = sprintf("$format", @arguments);
